/*
 *  map.cpp
 *
 *  Copyright (C) 2003-2004  Bruce Pennypacker (bruce@pennypacker.org)
 *
 *  This code is freely distributed under the GNU General Public
 *  License (http://www.gnu.org).  See COPYING for details.
 *
 *  So what does this code do?  It demonstrates a simple method of
 *  dynamically changing the colors on an image (in this case a map of 
 *  the USA) so that custom maps can be displayed in web browsers.  The
 *  way it works is very simple.  It uses the gd graphics manipulation
 *  library available at http://www.boutell.com/gd/ to load a PNG
 *  file into memory, manipulate the color palette of the image, then
 *  write it to stdout so a web browser can render it.
 *
 *  The trick that makes it all work is that each state in the map
 *  is drawn with a different RGB value.  This program simply changes
 *  each RGB value as desired.
 *
 *  I originally made this program specifically for the purpose of
 *  donating it to the Wheres George? website -
 *  http://www.wheresgeorge.com.  After a few recent requests I
 *  also decided to release the code to the public at large.
 *  (that means you!)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gd.h>

typedef struct tag_colormap {
  char *pState;
  int nBlue;
} COLORMAP, *PCOLORMAP;

#define COLORMAP_SIZE 52

/* The following table is the critical guts of the entire maping program.
 * If you edit usa_map.png you'll find that each state is a different
 * color (although it's not visible to the human eye).  Each state
 * looks red because the red value of each states RGB color is 222 and
 * the green value is 0.  (These are arbitrary values I picked)  The
 * blue value of each state is the value identified in the table below.
 * So the RGB value in usa_map.png of Alaska is 222:0:1, Alabama is
 * 222:0:2, Arkansas is 222:0:3, etc.  This program uses those values
 * to figure out what color to change each state to.
 */

const COLORMAP map[COLORMAP_SIZE] = {
  "AK", 1,
  "AL", 2,
  "AR", 3,
  "AZ", 4,
  "CA", 5,
  "CO", 6,
  "CT", 7,
  "DC", 8,
  "DE", 9,
  "FL", 10,
  "GA", 11,
  "HI", 12,
  "IA", 13,
  "ID", 14,
  "IL", 15,
  "IN", 16,
  "KS", 17,
  "KY", 18,
  "LA", 19,
  "MA", 20,
  "ME", 21,
  "MD", 22,
  "MI", 23,
  "MN", 24,
  "MO", 25,
  "MS", 26,
  "MT", 27,
  "NC", 28,
  "ND", 29,
  "NE", 30,
  "NH", 31,
  "NJ", 32,
  "NM", 33,
  "NV", 34,
  "NY", 35,
  "OH", 36,
  "OK", 37,
  "OR", 38,
  "PA", 39,
  "PR", 40,
  "RI", 41,
  "SC", 42,
  "SD", 43,
  "TN", 44,
  "TX", 45,
  "UT", 46,
  "VA", 47,
  "VT", 48,
  "WA", 49,
  "WI", 50,
  "WV", 51,
  "WY", 52
};


/*
 * Given a string in the format of "1:2:3" return the numerical
 * values 1, 2, and 3.  Make sure the values are in the range
 * of 0 and 255.
 */
int parseColor(const char *pColorString, int *r, int *g, int *b)
{
char pColor[80];
char *p, *q;

  memset(pColor, 0, sizeof(pColor));
  strncpy(pColor, pColorString, sizeof(pColor) - 1);

  if ((p = strchr(pColor, ':')) == NULL) {
    return (0);
  }

  *p = '\0';
  q = ++p;
  *r = atoi(pColor);

  if ((p = strchr(q, ':')) == NULL) {
    return (0);
  }

  *p = '\0';
  p++;

  *g = atoi(q);
  *b = atoi(p);

  if ((*r < 0) || (*r > 255) || (*g < 0) || (*g > 255) || (*b < 0) || (*b > 255)) {
    return (0);
  }

  return (1);
}


int main(int argc, char* argv[])
{

  const char *pFile = argv[1];
  const char *pTextColor = argv[2];
  const char *pDefColor = argv[3];

  int rHL, gHL, bHL;
  int rDef, gDef, bDef;
  int rTxt, gTxt, bTxt;

  int i, j, r, g, b;
  int color_index;
  char szStateList[256];

  if ((argc < 6) || ((argc - 4) % 2)) {
      fprintf(stderr, "Usage: map pngfile text_color default_color color state[:state....] [color state[:state...]] ...\n");
      exit(0);
  }


  FILE *fp;
  
  if ((fp = fopen(pFile, "rb")) == NULL) {
      fprintf(stderr, "Error: Unable to open %s for reading\n", pFile);
      exit(-1);
  }

  gdImagePtr im;

  im = gdImageCreateFromPng(fp);
  fclose(fp);

  if (!im) {
      fprintf(stderr, "Error: Unable to read PNG file %s\n", pFile);
      exit(-1);
  }

  if (!parseColor(pDefColor, &rDef, &gDef, &bDef)) {
      fprintf(stderr, "Error: invalid color %s\n", pDefColor);
      exit(-1);
  }

  if (!parseColor(pTextColor, &rTxt, &gTxt, &bTxt)) {
      fprintf(stderr, "Error: invalid color %s\n", pTextColor);
      exit(-1);
  }

  if ((color_index = gdImageColorExact(im, 0, 0, 255)) != -1) {
      im->red[color_index] = rTxt;
      im->green[color_index] = gTxt;
      im->blue[color_index] = bTxt;
  }

  for (i = 4 ; i < argc ; i += 2) {
    const char *pColor = argv[i];
    const char *pStateList = argv[i+1];
    memset(szStateList, 0, sizeof(szStateList));
    strncpy(szStateList, pStateList, sizeof(szStateList) - 1);

    if (!parseColor(pColor, &r, &g, &b)) {
        fprintf(stderr, "Error: invalid color %s\n", pDefColor);
        exit(-1);
    }

    char *pTok = strtok(szStateList, ":");

    while (pTok) {
      // Yeah, this could probably be a little faster as a binary search instead of a dumb for loop...
      for (j = 0 ; j < COLORMAP_SIZE; j++) {
          if (strcmp(map[j].pState, pTok) == 0) {
            if ((color_index = gdImageColorExact(im, 222, 0, map[j].nBlue)) == -1)
                break;

            im->red[color_index] = r;
            im->green[color_index] = g;
            im->blue[color_index] = b;
            break;
          }

      } // for (j)

      pTok = strtok(NULL, ":");

    } // while (pTok)

  } // for (i)


  // Set any colors that we haven't already modified to the default color
  for (i = 0 ; i < COLORMAP_SIZE; i++) {
    if ((color_index = gdImageColorExact(im, 222, 0, map[i].nBlue)) == -1)
        continue;

    im->red[color_index] = rDef;
    im->green[color_index] = gDef;
    im->blue[color_index] = bDef;
  }


  // Make white (the background color) transparent
  gdImageColorTransparent(im, gdImageColorExact(im, 255, 255, 255));

  gdImagePng(im, stdout);

  gdImageDestroy(im);

}

